#![no_std]
use core::cell::UnsafeCell;
use core::ops::{Deref, DerefMut, Drop};
use cortex_m::interrupt;

/// Internal replacement for `static mut T`
// TODO: Decide name and location.
pub struct RacyCell<T>(UnsafeCell<T>);

impl<T> RacyCell<T> {
    /// Create a RacyCell
    pub const fn new(value: T) -> Self {
        RacyCell(UnsafeCell::new(value))
    }

    /// Get &mut T
    pub unsafe fn get_mut_unchecked(&self) -> &mut T {
        &mut *self.0.get()
    }
}

// The type wrapped need to be Sync for RacyCell<T> to be Sync
//unsafe impl<T> Sync for RacyCell<T> where T: Sync {}
unsafe impl<T> Sync for RacyCell<T> {}

#[inline]
pub fn interrupt_free<F, R>(f: F) -> R
where
    F: FnOnce() -> R,
{
    if cfg!(target_arch = "arm") {
        interrupt::free(|_| f())
    } else {
        f()
    }
}

#[derive(Debug)]
pub enum Error {
    Uninitialized,
    OutOfMemory,
    AlreadyInitialized,
}

pub struct Box<T: 'static + Sized> {
    index: usize,
    p: &'static dyn P<T>,
}

impl<T: 'static> Drop for Box<T> {
    fn drop(&mut self) {
        self.p.dealloc(self.index)
    }
}

impl<T: 'static> Deref for Box<T> {
    type Target = T;
    fn deref(&self) -> &T {
        self.p.deref_index(self.index)
    }
}

impl<T: 'static> DerefMut for Box<T> {
    fn deref_mut(&mut self) -> &mut T {
        self.p.deref_mut_index(self.index)
    }
}

pub struct PoolMan<T: 'static, const S: usize> {
    data: RacyCell<[T; S]>,
    free: RacyCell<[usize; S]>,
    head: RacyCell<usize>,
    init: RacyCell<bool>,
}

pub trait P<T> {
    fn deref_index(&self, index: usize) -> &T;
    fn deref_mut_index(&self, index: usize) -> &mut T;
    fn dealloc(&self, index: usize);
}

impl<T: 'static, const S: usize> P<T> for PoolMan<T, S> {
    #[inline]
    fn deref_index(&self, index: usize) -> &T {
        unsafe {
            self.data /* RacyCell */
                .get_mut_unchecked() /* Array */
                .get_unchecked(index)
        }
    }

    #[inline]
    fn deref_mut_index(&self, index: usize) -> &mut T {
        unsafe {
            self.data /* RacyCell */
                .get_mut_unchecked() /* Array */
                .get_unchecked_mut(index)
        }
    }

    // Safety:
    //  dealloc can only be called by dropping Box
    //  which can happen only if already initialized
    //  thus it is safe to assume init == true
    //
    // not accissible from user code
    #[inline]
    fn dealloc(&self, index: usize) {
        interrupt_free(|| unsafe {
            *self.free.get_mut_unchecked().get_unchecked_mut(index) =
                *self.head.get_mut_unchecked();
            *self.head.get_mut_unchecked() = index;
        });
    }
}

unsafe impl<T: 'static, const S: usize> Sync for PoolMan<T, S> {}
unsafe impl<T: 'static, const S: usize> Send for PoolMan<T, S> {}

impl<T, const S: usize> PoolMan<T, S>
where
    T: Sized,
{
    pub const fn new(v: [T; S]) -> Self
    where
        T: Sized,
    {
        Self {
            data: RacyCell::new(v),
            free: RacyCell::new([0; S]),
            head: RacyCell::new(0),
            init: RacyCell::new(false),
        }
    }

    #[inline]
    pub fn init(&self) -> Result<(), Error> {
        interrupt_free(|| {
            if *unsafe { self.init.get_mut_unchecked() } {
                Err(Error::AlreadyInitialized)
            } else {
                let free = unsafe { self.free.get_mut_unchecked() };
                for (index, value) in free.iter_mut().enumerate() {
                    *value = index + 1;
                }
                *unsafe { self.init.get_mut_unchecked() } = true;
                Ok(())
            }
        })
    }

    #[inline]
    pub fn alloc(&'static self) -> Result<Box<T>, Error> {
        // this check does not need to be inside of critical section
        // as initialization is monotonic, cannot be undone
        if !unsafe { *self.init.get_mut_unchecked() } {
            Err(Error::Uninitialized)
        } else {
            interrupt_free(|| unsafe {
                // head of free list
                let head = *self.head.get_mut_unchecked();
                if head >= S {
                    Err(Error::OutOfMemory)
                } else {
                    // update head to `next` free
                    *self.head.get_mut_unchecked() =
                        *self.free.get_mut_unchecked().get_unchecked(head);
                    Ok(Box {
                        index: head,
                        // allocator: self,
                        p: self,
                    })
                }
            })
        }
    }
}
